"""Created by Purushot on 25/11/18"""

import string
from random import randrange

__author__ = "Purushot14"

from .snowflake import Snowflake

# This original was ordered by ascii code, here `-` was 45, digits are between 48 and 57,
# upper alphabets were between 65 and 90, `_` was 95 and lower alphabets were between 97 and 122
ORIGINAL = "-" + string.digits + string.ascii_uppercase + "_" + string.ascii_lowercase

SHUFFLED = None
snowflake = Snowflake()


def _shuffle():
    """:return:"""
    source_str = ORIGINAL
    target_str = ""

    while source_str.__len__() > 0:
        r = randrange(0, source_str.__len__())
        char = source_str[r]
        target_str += char
        source_str = source_str.replace(char, "", 1)
    return target_str


def _get_shuffled():
    global SHUFFLED
    if SHUFFLED:
        return SHUFFLED
    SHUFFLED = _shuffle()
    return SHUFFLED


def _int_to_base64(num: int, is_ordered=True):
    """Converts a positive integer into a base36 string."""
    assert num >= 0
    alphabets = ORIGINAL if is_ordered else _get_shuffled()

    res = ""
    while not res or num > 0:
        num, i = divmod(num, 63)
        res = alphabets[i] + res
    return res


def get_next_snowflake_id(mult=None):
    global snowflake
    if mult:
        snowflake.set_mult(mult)
    return snowflake.__next__()


def generate_short_id(mult=None, is_ordered=True):
    _snowflake_id = get_next_snowflake_id(mult)
    return _int_to_base64(_snowflake_id, is_ordered)



"""Created by Purushot at 30/11/22"""

__author__ = "Purushot14"

import ipaddress
import logging
import os
import time
import uuid
from _socket import gaierror
from socket import gethostbyname, gethostname
from threading import Lock

# Wednesday, 1 January 2020 00:00:00 GMT
start_epoch = 1577817000
machine_id_bits = 16
process_id_bits = 8
sequence_bits = 5
machine_id_mask = -1 ^ (-1 << machine_id_bits)
process_id_mask = -1 ^ (-1 << process_id_bits)
sequence_mask = -1 ^ (-1 << sequence_bits)
timestamp_left_shift = sequence_bits + process_id_bits + machine_id_bits


def get_machine_id(ip_address: str = None, is_retry=False):
    try:
        ip_address = ip_address or gethostbyname(gethostname() or "localhost")
    except gaierror:
        if is_retry:
            return uuid.getnode() & machine_id_mask
        return get_machine_id(is_retry=True)
    return int(ipaddress.ip_address(ip_address)) & machine_id_mask


def get_process_id():
    return os.getpid() & process_id_mask


def get_worker_id():
    return get_machine_id() | get_process_id()


class Snowflake:
    def __init__(self, worker_id=None, mult=10000):
        logging.info("Snowflake init called")
        self.worker_id = worker_id or get_worker_id()
        self._mult = mult
        self.__last_timestamp = -1
        self.__sequence = 0
        self._lock = Lock()

    def set_mult(self, mult):
        if isinstance(mult, int) and mult > 0:
            self._mult = mult
            time.sleep(1 / mult)
        else:
            raise ValueError("mult must be a positive integer")

    @property
    def start_epoch(self):
        return start_epoch * self._mult

    @property
    def current_time(self) -> int:
        return int(self._mult * time.time())

    def _next_id(self):
        current_time = self.current_time
        if self.__last_timestamp == current_time:
            self.__sequence = (self.__sequence + 1) & sequence_mask
            if self.__sequence == 0:
                time.sleep(1 / self._mult)
                self.__sequence = 0
                return self._next_id()
        else:
            self.__sequence = 0
        self.__last_timestamp = current_time
        sequence = self.__sequence << machine_id_bits
        return ((current_time - self.start_epoch) << timestamp_left_shift) | self.worker_id | sequence

    def get_next_id(self):
        with self._lock:
            return self._next_id()

    def __next__(self):
        return self.get_next_id()

    def __iter__(self):
        return self
